/******************************************************************
 * Railcom (Bi-Di) detection and display
 * Version: 20201217
 * Copyright 2021- DesktopStation, Nagoden
 * Wiki: https://desktopstation.net/wiki/doku.php/railcomdisplay
 ******************************************************************/

#include <arduino.h>
#include "sseg.h"

//#define DEBUG

#define TIME_RECVWAIT 25
#define SIZE_RECV 16
#define SIZE_CONV 8


#define CHK_RCID2 B0010 //アドレスLow
#define CHK_RCID1 B0001 //アドレスHigh

typedef struct {
  byte mID;
  byte mData1; //12bit
  byte mData2; //18bit
  byte mData3; //24bit / 12bit + 12bit
  byte mData4; //36bit  
}
RAILCOM_DATAGRAM;

//Codings in the 4 out of 8 code NMRA RCN-217 Table2
//変換テーブル
const int convtable[256] = { 
255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 
255, 255, 255, 255, 255, 255, 51, 255, 255, 255, 52, 255, 53, 54, 255, 255, 255, 255, 
255, 255, 255, 255, 58, 255, 255, 255, 59, 255, 60, 55, 255, 255, 255, 255, 63, 255, 
61, 56, 255, 255, 62, 57, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 
36, 255, 255, 255, 35, 255, 34, 33, 255, 255, 255, 255, 31, 255, 30, 32, 255, 255, 29, 
28, 255, 27, 255, 255, 255, 255, 255, 255, 25, 255, 24, 26, 255, 255, 23, 22, 255, 21, 
255, 255, 255, 255, 37, 20, 255, 19, 255, 255, 255, 50, 255, 255, 255, 255, 255, 255, 
255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 14, 255, 13, 12, 255, 255, 
255, 255, 10, 255, 9, 11, 255, 255, 8, 7, 255, 6, 255, 255, 255, 255, 255, 255, 4, 255, 
3, 5, 255, 255, 2, 1, 255, 0, 255, 255, 255, 255, 15, 16, 255, 17, 255, 255, 255, 18, 
255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 43, 48, 255, 255, 42, 47, 
255, 49, 255, 255, 255, 255, 41, 46, 255, 45, 255, 255, 255, 44, 255, 255, 255, 255, 
255, 255, 255, 255, 255, 40, 255, 39, 255, 255, 255, 38, 255, 255, 255, 255, 255, 255, 
255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 };


// Pins definitions 7 segments
SSEG SegDisplay;

//DEBUG
unsigned long gPreviousL1 = 0; // 2ms interval

//表示アドレス
uint16_t locoadress;
word gAddrDataLow = 0;
word gAddrDataHigh = 0;
word gAddrReceived = 0;


//制御用変数
uint8_t interruptflag = 0;
long no_signaltime = 0;
int no_recive = 0;
long no_commandtime = 0;
int no_command = 0;
byte gWaitTime = 0;//CutOutからの受信ウェイト用(50ms)

//受信データの変数
byte receivedByte = B00000000;
int command_count = 0;
byte gReciveData[SIZE_RECV] = {
  0
};
byte gConversionData[SIZE_CONV] = {
  0
};


byte bitwork1 = B00000000;
byte bitwork2 = B00000000;
byte bitworkA = B00000000;
byte bitworkB = B00000000;

RAILCOM_DATAGRAM gRecvDataGram[SIZE_CONV];
 	
void clearAddr();


void setup() {

  //7SEG初期化
  SegDisplay.SetInit();

  //Display Clear
  SegDisplay.SetNumber(0);

  //シリアル通信開始
  Serial.begin(250000);

  for (int i = 0; i < 3; i++) {
    gRecvDataGram[i].mID = 0;
    gRecvDataGram[i].mData1 = 0;
    gRecvDataGram[i].mData2 = 0;
    gRecvDataGram[i].mData3 = 0;
    gRecvDataGram[i].mData4 = 0;
  }

  //Task
  gPreviousL1 = millis();

}

void loop() 
{
  //カットアウト信号がlowになったらinterruptflagを１にしてシリアルポートから受信
  if (((PIND & _BV(2)) == LOW) && (interruptflag == 0)) {
    
    interruptflag = 1;

    //受信ウェイト(RailComの信号が溜まるまで待つ、通常は0.5msで終了。)
    gWaitTime = TIME_RECVWAIT;
    
    #ifdef DEBUG
    Serial.println("!CUTOUT!");
    #endif
  }

  //Railcom Decode
  if(gWaitTime == 0)
  {
    if (interruptflag == 1) {
      // シリアルポートにデータがいるかチェック
      if (Serial.available()) {
        
	    #ifdef DEBUG
	    Serial.println("!RECV!");
	    #endif
        receiveSerial();

      } else {
        //CutOUTがDisableになった場合に2秒で表示をリセットします。
        if (no_recive == 0) {
          no_signaltime = millis();
          no_recive = 1;
        } else if ((millis() - no_signaltime) >= 2000) {
          command_count = 0;
          interruptflag = 0;
          no_recive = 0;
           clearAddr();
        	

          //Display Clear
          SegDisplay.SetNumber(0);
        }
      }
    }
  }

  //Display
	if( (millis() - gPreviousL1) >= 2)
	{
		//Reset task
		gPreviousL1 = millis();
		
    //Display 
    SegDisplay.ShowNumber();

    //WaitTime
    if( gWaitTime > 0 )
    {
      gWaitTime--;
    }
	}
}

void receiveSerial()
{
  no_recive = 0;

  while( Serial.available() > 0 )
  {
      //データ受信
      receivedByte = Serial.read();

      //データが00hは読み取りません 0x0fはNACKコードなのでこれも無視します
      if ((receivedByte != 0) && (receivedByte != 0x0f)) {
        gReciveData[command_count] = receivedByte;
        command_count = command_count + 1;
        no_command = 0;
      }
      //データが00hで2秒連続した場合は表示をリセットします。 
      else if (receivedByte == 0) {
        if (no_command == 0) 
        {
          no_commandtime = millis();
          no_command = 1;
        } else if ((millis() - no_commandtime) >= 2000) 
        {
          command_count = 0;
          interruptflag = 0;
          no_command = 0;
           clearAddr();
        	
          //Display Clear
          SegDisplay.SetNumber(0);
        	
        	return;
        }
      }
  }

  //Clear
  //Serial.flush();
  

  //所定のバイト数受信したら受信データを処理します
  //アドレスは2バイトのデータを2回に分けて送信してくるので4個受信します
  if (command_count >= 4) {

    #ifdef DEBUG
      Serial.print(" ReceivedByte ");
    for (int i = 0; i < SIZE_RECV; i++) {
      Serial.print(gReciveData[i], HEX);
      Serial.print(" ");
    }
    #endif
      Serial.println("");

    //受信したコマンドを変換テーブルで変換 
    excec_conversion(command_count);

    //デバッグ用
    #ifdef DEBUG
    for (int i = 0; i < SIZE_CONV; i++) {
      //Serial.print(i);
      //Serial.print(" ConvertData ");
      //Serial.println(gConversionData[i], HEX);
      //Serial.println(" ");
    }
    #endif
    //受信データクリア            
    for (int i = 0; i < SIZE_RECV; i++) {
      gReciveData[i] = 0;
    }

  }
    //フラグ変数クリア　
    command_count = 0;
    interruptflag = 0;

}

void excec_conversion(byte inRecvSize) {

  byte aErrChk = 0;
  byte aDblSendChk = 0;
  byte aLen = 2;//inRecvSize / 2;

  //受信したデータを変換テーブルで変換
  for (int i = 0; i < inRecvSize; i++) {
    gConversionData[i] = convtable[gReciveData[i]];

    //4/8 codingに合っていない形式の場合はエラーなので、フラグを立てる
    if (gConversionData[i] == 255) {
      aErrChk++;
    }

  }

  if (aErrChk == 0) {

    //変換した値を下位12ビット分で　1番目と2番目　3番目と4番目で結合
    // 1番目00010110 2番目 00101100 -> 010110 + 101100 -> 010110101100
    //work変数のクリア
    bitwork1 = B00000000;
    bitwork2 = B00000000;
    bitworkA = B00000000;
    bitworkB = B00000000;

    //結合した値を上位4ビットと下位4ビットに分割
    //2桁アドレス
    //1番目と2番目0010.00000011  <-下位8ビットがCV03(2桁アドレス）
    //3番目と4番目0001.00000000  <-上位4ビットが0001で下位8ビットが0であれば2桁アドレス　
    //4桁アドレス
    //1番目と2番目0001.10001110  <-上位4ビットが0001で下位8ビットがCV17
    //3番目と4番目0010.11110110  <-上位4ビットが0010で下位8ビットがCV18
    for (int i = 0; i < aLen; i++) {

      bitwork1 = gConversionData[(i * 2)];
      bitwork2 = gConversionData[(i * 2) + 1];

      //bitwork1　下位2BitをbitworkAの上位2ビットに移動　（ローテート命令があるといいのに）
      bitWrite(bitworkA, 0, bitRead(bitwork1, 0));
      bitWrite(bitworkA, 1, bitRead(bitwork1, 1));
      bitworkA = bitworkA << 6;
      //上位2ビットでマスクして他のビットをクリア
      bitworkA = bitworkA & B11000000;
      // bitwork2の下位6ビットをマスクし、bitworkA,Bを合成（下位バイトのデータ）
      bitworkB = bitwork2 & B00111111;
      bitworkB = bitworkB |= bitworkA; //下の値(アドレス値）
      //bitwork1の値を下位方向に2ビットシフトして下位4ビットでマスク（上位バイトのデータ）
      bitworkA = bitwork1 >> 2;
      bitworkA = bitworkA & B00001111; //上の値（ID)

      //一旦配列に保持する
      gRecvDataGram[i].mID = bitworkA;
      gRecvDataGram[i].mData1 = bitworkB;

      //IDごとに処理する
      switch (gRecvDataGram[i].mID) {
        //case 0:
        //POM
        //break;
      case 2:
        //low addr

        if ((gAddrReceived & CHK_RCID2) > 0) {
          //同じIDが複数回来ても一致していればOK、違うならパケット異常

          if( gAddrDataLow != gRecvDataGram[i].mData1)
          {
           clearAddr();
          	
            //デバッグ用
            #ifdef DEBUG
            Serial.println("Double check error!");
            #endif
          }
        }
        else
        {
          gAddrDataLow = gRecvDataGram[i].mData1;
          gAddrReceived |= CHK_RCID2;
        }

        break;
      case 1:
        //high addr

        if ((gAddrReceived & CHK_RCID1) > 0) {
          //同じIDが複数回来ても一致していればOK、違うならパケット異常

          if( gAddrDataHigh != gRecvDataGram[i].mData1)
          {
            clearAddr();
            //デバッグ用
            #ifdef DEBUG
            Serial.println("Double check error!");
            #endif
          }     

        } else {
          gAddrDataHigh = gRecvDataGram[i].mData1;
          gAddrReceived |= CHK_RCID1;
        }

        break;
      default:
        //その他のID
        break;
      }

      //デバッグ用
      #ifdef DEBUG
      Serial.print("id=");
      Serial.println(gRecvDataGram[i].mID);
      Serial.print("Data=");
      Serial.println(gRecvDataGram[i].mData1);
      #endif
    }

    //アドレスの取得
  	if( gAddrReceived == 3)
  	{
	    if ((gAddrDataLow != 0) && (gAddrDataHigh != 0)) {
	      //4digit 
	      locoadress = ((gAddrDataHigh & 63) * 256) + gAddrDataLow;

	    } else if (gAddrDataLow != 0) {
	      //2digit 
	      locoadress = gAddrDataLow;
	    } else {
	      //2digit 
	      locoadress = 0;
	    }
  	}
  	else
  	{
  		locoadress = 0;
  	}
    //デバッグ用
    #ifdef DEBUG
    Serial.print(" LocoAddress ");
    Serial.println(locoadress);
    #endif

    if (locoadress > 0) {
      //LEDに数字表示
      SegDisplay.SetNumber(locoadress);
    }

  }
}
 	
void clearAddr()
{
    gAddrDataLow = 0;
    gAddrDataHigh = 0;
  	gAddrReceived = 0;
	
	//Clear
    for (int i = 0; i < SIZE_CONV; i++) 
	{
      gRecvDataGram[i].mID = 0;
      gRecvDataGram[i].mData1 = 0;
    }
	
}